/* Copyright (c) 2018 Griefer@Work                                            *
 *                                                                            *
 * This software is provided 'as-is', without any express or implied          *
 * warranty. In no event will the authors be held liable for any damages      *
 * arising from the use of this software.                                     *
 *                                                                            *
 * Permission is granted to anyone to use this software for any purpose,      *
 * including commercial applications, and to alter it and redistribute it     *
 * freely, subject to the following restrictions:                             *
 *                                                                            *
 * 1. The origin of this software must not be misrepresented; you must not    *
 *    claim that you wrote the original software. If you use this software    *
 *    in a product, an acknowledgement in the product documentation would be  *
 *    appreciated but is not required.                                        *
 * 2. Altered source versions must be plainly marked as such, and must not be *
 *    misrepresented as being the original software.                          *
 * 3. This notice may not be removed or altered from any source distribution. *
 */
#ifndef GUARD_KERNEL_INCLUDE_I386_KOS_PIC_H
#define GUARD_KERNEL_INCLUDE_I386_KOS_PIC_H 1

#include <hybrid/compiler.h>
#include <kos/types.h>
#include <sys/io.h>
#include "interrupt.h"

DECL_BEGIN

/* PIC (Programmable Interrupt Controller) API. */
#define X86_PIC1       0x20 /* IO base address for master PIC */
#define X86_PIC2       0xa0 /* IO base address for slave PIC */
#define X86_PIC1_CMD   X86_PIC1
#define X86_PIC1_DATA (X86_PIC1+1)
#define X86_PIC2_CMD   X86_PIC2
#define X86_PIC2_DATA (X86_PIC2+1)

#define X86_PIC_CMD_EOI     0x20 /* End-of-interrupt command code */
#define X86_ICW1_ICW4       0x01 /* ICW4 (not) needed */
#define X86_ICW1_SINGLE     0x02 /* Single (cascade) mode */
#define X86_ICW1_INTERVAL4  0x04 /* Call address interval 4 (8) */
#define X86_ICW1_LEVEL      0x08 /* Level triggered (edge) mode */
#define X86_ICW1_INIT       0x10 /* Initialization - required! */
#define X86_ICW4_8086       0x01 /* 8086/88 (MCS-80/85) mode */
#define X86_ICW4_AUTO       0x02 /* Auto (normal) EOI */
#define X86_ICW4_BUF_SLAVE  0x08 /* Buffered mode/slave */
#define X86_ICW4_BUF_MASTER 0x0c /* Buffered mode/master */
#define X86_ICW4_SFNM       0x10 /* Special fully nested (not) */
#define X86_PIC_READ_IRR    0x0a /* OCW3 irq ready next CMD read */
#define X86_PIC_READ_ISR    0x0b /* OCW3 irq service next CMD read */

#ifdef __CC__
#define X86_PIC1_IRR()  (outb(X86_PIC1_CMD,X86_PIC_READ_IRR),inb(X86_PIC2_CMD))
#define X86_PIC2_IRR()  (outb(X86_PIC2_CMD,X86_PIC_READ_IRR),inb(X86_PIC2_CMD))
#define X86_PIC1_ISR()  (outb(X86_PIC1_CMD,X86_PIC_READ_ISR),inb(X86_PIC2_CMD))
#define X86_PIC2_ISR()  (outb(X86_PIC2_CMD,X86_PIC_READ_ISR),inb(X86_PIC2_CMD))

/* Signal EOI (End of interrupt) to the first (master), or second (slave) PIC. */
#define X86_PIC_EOI(intno) ((intno) >= X86_INTERRUPT_PIC2_BASE ? outb(X86_PIC2_CMD,X86_PIC_CMD_EOI) : (void)0, \
                             outb(X86_PIC1_CMD,X86_PIC_CMD_EOI))
#define X86_PIC1_EOI()      (outb(X86_PIC1_CMD,X86_PIC_CMD_EOI))
#define X86_PIC2_EOI()      (outb(X86_PIC2_CMD,X86_PIC_CMD_EOI),outb(X86_PIC1_CMD,X86_PIC_CMD_EOI))


struct x86_spurious_interrupts {
    u32 sp_pic1; /* Amount of spurious interrupts that occurred on PIC #1. */
    u32 sp_pic2; /* Amount of spurious interrupts that occurred on PIC #2. */
    u32 sp_apic; /* Amount of spurious interrupts that occurred through the APIC. */
};

/* Tracking information about spurious interrupts.
 * >> Useful for detecting faulty software that sends EOI commands at the
 *    wrong time, or for tracking problems in hardware, such as line noise. */
DATDEF ATTR_PERCPU struct x86_spurious_interrupts x86_spurious_interrupts;


/* Get/Set the mask of disabled interrupt lines.
 * >> `INTNO_PIC1_STMASK()' disables irq_t: 0x20...0x27
 * >> `INTNO_PIC2_STMASK()' disables irq_t: 0x28...0x2f */
#define X86_PIC1_GTMASK()   inb_p(X86_PIC1_DATA)
#define X86_PIC1_STMASK(m) outb_p(X86_PIC1_DATA,m)
#define X86_PIC2_GTMASK()   inb_p(X86_PIC2_DATA)
#define X86_PIC2_STMASK(m) outb_p(X86_PIC2_DATA,m)
#endif /* __CC__ */


/* Standard ISA IRQ numbers.
 * >> This is the default wireing of hardware interrupts. */
#define X86_INTNO_PIC1_PIT   X86_INTNO_PIC1(0) /* Programmable Interrupt Timer Interrupt. */
#define X86_INTNO_PIC1_KBD   X86_INTNO_PIC1(1) /* Keyboard Interrupt. */
#define X86_INTNO_PIC1_CAS   X86_INTNO_PIC1(2) /* Cascade (used internally by the two PICs. never raised). */
#define X86_INTNO_PIC1_COM2  X86_INTNO_PIC1(3) /* COM2 (if enabled). */
#define X86_INTNO_PIC1_COM1  X86_INTNO_PIC1(4) /* COM1 (if enabled). */
#define X86_INTNO_PIC1_LPT2  X86_INTNO_PIC1(5) /* LPT2 (if enabled). */
#define X86_INTNO_PIC1_FLOP  X86_INTNO_PIC1(6) /* Floppy Disk. */
#define X86_INTNO_PIC1_LPT1  X86_INTNO_PIC1(7) /* LPT1 / Unreliable "spurious" interrupt (usually). */
#define X86_INTNO_PIC2_CMOS  X86_INTNO_PIC2(0) /* CMOS real-time clock (if enabled). */
#define X86_INTNO_PIC2_FREE1 X86_INTNO_PIC2(1) /* Free for peripherals / legacy SCSI / NIC. */
#define X86_INTNO_PIC2_FREE2 X86_INTNO_PIC2(2) /* Free for peripherals / SCSI / NIC. */
#define X86_INTNO_PIC2_FREE3 X86_INTNO_PIC2(3) /* Free for peripherals / SCSI / NIC. */
#define X86_INTNO_PIC2_PS2M  X86_INTNO_PIC2(4) /* PS2 Mouse. */
#define X86_INTNO_PIC2_FPU   X86_INTNO_PIC2(5) /* FPU / Coprocessor / Inter-processor. */
#define X86_INTNO_PIC2_ATA1  X86_INTNO_PIC2(6) /* Primary ATA Hard Disk. */
#define X86_INTNO_PIC2_ATA2  X86_INTNO_PIC2(7) /* Secondary ATA Hard Disk. */

DECL_END

#endif /* !GUARD_KERNEL_INCLUDE_I386_KOS_PIC_H */
